#!/usr/bin/bash

#-----------------------------------------------------------
# Simple chat written as bash script set of four scripts:
# chat.sh,
# connect.sh,
# usuc.sh
# suc.sh
# See https://the-dam.org/docs/explanations/SecureBash.html
#-----------------------------------------------------------
#
# Encryption/decryption howto:
#
# Imagine you want to encrypt the string “your very secret stuff“, you can achieve this 
# by issuing the following command:
# $ echo -n 'your very secret stuff' | openssl enc -aes128 -pbkdf2 -A -a -e -k some-password
# $ U2FsdGVkX18x/szJ03GI3RMuISEwu+RYRumtuO2ABLHmaPLXQgyOvdL7GDcw7ouc
#
# In this example we are:
#
#     echoing the “your very secret stuff” without the trailing newline (-n).
#    openssl enc: calling the symmetric cipher routines of the openssl program.
#    -aes128: choosing AES-128 cipher.
#    -pbkdf2: using PBKDF2 algorithm to derive the key from the password.
#    -A: write result in a single line
#    -a: if encryption is taking place the data is base64 encoded after encryption. 
#        If decryption is set then the input data is base64 decoded before being decrypted.
#    -e: choosing to encrypt the input data, or -d to decrypt
#    -k: some-password: setting the password “some-password”. This password will be asked 
#        at the time of decrypting the data. Additionally, if we don’t provide it as an argument now, we’ll be prompted for it.
#
# The process to decrypt the data is almost identical. You’d just need to replace the -e flag (encrypt) with -d (decrypt):
# $ echo 'U2FsdGVkX18fl6CDdLoTthdNlfIU4LcYwa4TGaxy/q6UNlOFl3G1tRC90l+oeOx6' | openssl enc -aes128 -pbkdf2 -a -d -k some-password
# $ Your very secret stuff
#
#---------------------------------------------------------------------------------

# Variable definition
# ================================================================================

set -euo pipefail;
#tailpidremote="";
NAME="";
PASSWORD="";
CHANNEL="";
CHOISE="";
LOGIN="";
TAILLINES="20";

# Force only the child script to stop when Ctrl+C is pressed - exit messages
# ==================================================================================

sigterm_handler() {
  echo
  if [[ $NAME != "" ]] && [[ $PASSWORD != "" ]] && [[ $CHANNEL != "" ]]  && [[ $LOGIN != "" ]]
  then
     echo -e "----- $(printf '\e[38;5;%dm\e[48;5;%dm%-9s\e[0m %s\e[0m\n' $FGC $BGC $NAME)------- left the chat -------" | openssl enc -aes128 -pbkdf2 -A -a -e -k "$PASSWORD" | tee -a ./chat/$CHANNEL >> "./chat/$CHANNEL.out";
     echo "" >> ./chat/$CHANNEL.out
     echo "" >> ./chat/$CHANNEL
     echo "--- EXIT > CHAT CLOSED ---"
     echo
  else
     echo "Login FAILURE. Closing the chat."
     echo
  fi
#  if [[ $tailpidremote != "" ]]
#  then
#     kill $tailpidremote 2>&1 > /dev/null;
#  fi
  if ls ./$NAME.*-fifo 1> /dev/null 2>&1;
  then
  rm ./$NAME.*-fifo # delete line fifo files in current directory if they exist
  fi
  if ls ./$NAME.*.status 1> /dev/null 2>&1;
  then
  rm ./$NAME.*.status # delete connection status files in current directory if they exist
  fi
  exit 1
}

trap 'trap " " SIGINT SIGTERM SIGHUP; kill 0; wait; sigterm_handler' SIGINT SIGTERM SIGHUP # Setup signal trap

# Initialisation
# ==========================================================================

clear # Clean the screan from previuos text

# Validate command line

if [ $# -gt 0 ]; then
    echo "Usage: $0"
    exit 1
fi

# Check and connect if exists, or create chat archive file name
# ==========================================================================

clear
echo "Connecting to \"bChat\" Private Chat server"
echo "---------------------------------------------"
echo
if [ -z "$( ls -A './chat' )" ]; then
   echo "No local channels created yet.";
else
   echo "Available local channels:"
   ls ./chat | sed -n 's/\.in$//p' | sed 's/^/ -/'
fi
echo
read -p "Choose local chat channel or create new: " CHOISE
printf "\033[A\33[2K\r" # move cursor up and clear the line
if [[ -z $CHOISE ]] # exit if chanel name is empty string
then
   echo "You did not provide channel name. Closing the chat."
   echo;
   exit 1;
fi
if [ -f ./chat/$CHOISE.in ] # Check if chat archive files exists
then
   CHANNEL=$CHOISE;
   touch ./chat/$CHOISE.out;
   touch ./chat/$CHOISE;
else
   read -p "Create new local chat channel \"$CHOISE\"? (y/n[n]): " ANSWER;
   printf "\033[A\33[2K\r" # move cursor up and clear the line
   case ${ANSWER:0:1} in
   y|Y )
   touch ./chat/$CHOISE.in;
   touch ./chat/$CHOISE.out;
   touch ./chat/$CHOISE;
   echo "Created channel \"$CHOISE\"";
   CHANNEL=$CHOISE;
   ;;
   * )
   echo "Try again to connect to local chat channels.";
   echo;
   exit 1;
   ;;
   esac
fi

# ------------------------------ START OF THE CHAT -----------------------------
# ==============================================================================

echo "Connected to \"$CHANNEL\" channel locally"
echo "---------------------------------------------"
echo "Press Ctrl+C to leave the channel"
echo "Type ~HISTORY~ for full chat history"
echo "Type ~CONNECT~ for remote connections"
echo "---------------------------------------------"
echo "!!! ATTENTION !!! Type ~DESTROY~"
echo "at any time by any user to safely"
echo "erase data and destroy the chat."
echo "---------------------------------------------"

# Assign user nickname
# =============================================================================

echo 
read -p "Enter your nickname: " NAME
if [[ $NAME == "" ]]
then
    echo
    echo "No nickname specified, press \"Enter\""
    echo "to be registered as a guest, or "
    read -p "enter your nickname: " NAME
    if [[ $NAME == "" ]]
    then
	echo
	echo "You are registered as \"Guest\"";
	NAME="Guest";
    fi
fi
echo

# Setting chat password
# =============================================================================

echo "Password should be at least 8 char."
read -s -p "Enter encryption password: " PASSWORD
echo -en "\033[1K\r"
read -s -p "Verify encryption password: " PASSWORD_V
echo -en "\033[1K\r"
if [[ $PASSWORD == $PASSWORD_V ]] # Check if passwords match
then
    if [[ $PASSWORD == "" ]] # Check if password is not empty
    then 
        echo "No password entered. Closing the chat."
        echo
        exit 1
    fi
    if [[ ${#PASSWORD} -lt 8 ]] # Check if password is at least 8 characters
    then 
        echo "Password too short. Closing the chat."
        echo
        exit 1
    fi
    printf "\033[A\33[2K\r" # move cursor up and clear the line
    echo "OK: Passwords MATCH";
    if [[ $(tail -n1 ./chat/$CHANNEL.out | openssl enc -aes128 -pbkdf2 -A -a -d -k $PASSWORD 2>&1 >/dev/null | grep "bad decrypt") == "" ]] # check if password is good
    then
       echo "Password ACCEPTED";
    else # exit if decryption of out file using submitted password gives en error
       sleep 5;
       echo "Password FAILURE. Closing the chat";
       echo
       exit 1
    fi 
else
    echo "Passwords DO NOT MATCH"
    read -s -p "Enter encryption password: " PASSWORD
    echo -en "\033[1K\r"
    read -s -p "Verify encryption password: " PASSWORD_V
    echo -en "\033[1K\r"
    if [[ $PASSWORD == $PASSWORD_V ]]
    then
        if [[ $PASSWORD == "" ]] # Check if password is empty
        then 
            echo "No password entered. Closing the chat.";
            echo;
            exit 1;
        fi
        if [[ ${#PASSWORD} -lt 8 ]] # Check if password is at least 8 characters
        then 
        echo "Password too short. Closing the chat.";
        echo;
        exit 1;
        fi
           printf "\033[A\33[2K\r" # move cursor up and clear the line
           echo "OK: Passwords MATCH";
        if [[ $(tail -n1 "./chat/$CHANNEL.out" | openssl enc -aes128 -pbkdf2 -A -a -d -k $PASSWORD 2>&1 >/dev/null | grep "bad decrypt") == "" ]] # check if password is good
        then
           echo "Password ACCEPTED";
        else
           sleep 5;
           echo "Password FAILURE. Closing the chat.";
           echo;
           exit 1;
        fi 
    else # pasword not maching twice, closing chat
        echo
        sleep 5;
        echo "Password FAILURE. Closing the chat.";
        echo;
        exit 1;
    fi
fi
FGC=$(( 16#$(\echo "$NAME" | /usr/bin/md5sum | /usr/bin/head -c2) ))  # Arbitrary number in [0:255]
BGC=$(( ("$FGC" + 128) % 255 ))  # Most of the time, high enough contrast
echo -e "                $(date)" | openssl enc -aes128 -pbkdf2 -A -a -e -k "$PASSWORD" | tee -a ./chat/$CHANNEL >> "./chat/$CHANNEL.out";
echo "" >> "./chat/$CHANNEL.out";
echo "" >> "./chat/$CHANNEL";
echo -e "----- $(printf '\e[38;5;%dm\e[48;5;%dm%-9s\e[0m %s\e[0m\n' $FGC $BGC $NAME)------ joined the chat ------" | openssl enc -aes128 -pbkdf2 -A -a -e -k "$PASSWORD" | tee -a ./chat/$CHANNEL >> "./chat/$CHANNEL.out";
echo "" >> "./chat/$CHANNEL.out";
echo "" >> "./chat/$CHANNEL";
LOGIN="success"

remote_connect()
{
# Connect to remote host and append each line from remote "channel.out" to local "channel.in" file
# ================================================================================================

echo "Connecting to remote chat hosts"
echo "---------------------------------------------"
echo
read -p "Connect to remote hosts ? (y/n[n]): " CONNECT;
printf "\033[A\33[2K\r" # move cursor up and clear the line
case ${CONNECT:0:1} in
   y|Y )
   while true; do    # Loop forever (until break is issued)
      read -p "Enter remote user@ipaddress: " REMOTE
      read -p "Enter remote port: " PORT
      read -p "Enter remote password: " PASS
      printf "\033[A\33[2K\r" # move cursor up and clear the line
      echo "Connecting..."
      REMDIR="~/Work/suc/bash-chat-duo-2"
      # running line appending external script
      if [ -f ./*.$REMOTE:$PORT.status ] # check if conection to remote host already exist
      then
         printf "\033[A\33[2K\r" # move cursor up and clear the line
         echo "This chat host is already CONNECTED" # do not connect if already connected to this remote chat host
      else
         nohup $SHELL -c "./connect.sh $REMOTE $PORT $PASS $REMDIR $CHANNEL 2>&1 | grep --line-buffered 'Authenticated\|cannot' > ./$NAME.$REMOTE:$PORT.status" > /dev/null 2>&1 &  # doesn't create nohup.out
         # tailpidremote=$!
         sleep 2
         if grep -q "Authenticated" ./$NAME.$REMOTE:$PORT.status; then
             printf "\033[A\33[2K\r" # move cursor up and clear the line
             echo "CONNECTED to \"$REMOTE:$PORT\" chat host"
             if grep -q "cannot" ./$NAME.$REMOTE:$PORT.status; then
                echo "Remote channel DOES NOT EXIST"
             fi
         else
         printf "\033[A\33[2K\r" # move cursor up and clear the line
         echo "Connection FAILED. Try again..."
         rm ./$NAME.$REMOTE:$PORT.status # delete connection status file
         fi
      fi
      read -p "Connect to another remote host? (y/n[n]):" ANOTHER
      printf "\033[A\33[2K\r" # move cursor up and clear the line
      case ${ANOTHER:0:1} in
          y|Y)
          :
          ;;
          *)
          if ls ./*.status 1> /dev/null 2>&1; # print "no connections" message if no connection status files exist
          then
          :
          else
          echo "No remote connections - local/ssh chat only";
          fi
          break
          ;;
      esac
   done
   ;;
   * )
   printf "\033[A\33[2K\r" # move cursor up and clear the line
   if ls ./*.status 1> /dev/null 2>&1; # print "no connections" message if no connection status files exist
   then
   :
   else
   echo "No remote connections - local/ssh chat only";
   fi
   echo;
   ;;
esac

echo
echo "Current CONNECTIONS:"
find -name "*.status" | cut -d '.' -f 3
echo
read -p "Press \"Enter\" to chat:"
printf "\033[A\33[2K\r" # move cursor up and clear the line
echo "---------------------------------------------"
echo
echo
}

# Run chat
# ===========

echo
echo "--- WELCOME TO THE CHAT ---"
echo
echo
while true
do # infinite loop to implement special command running in chat
  ./usuc.sh "$CHANNEL" "$NAME" "$PASSWORD" "$TAILLINES" # run user chat script with exit codes depending on entered words
  if grep -q Connecting ./usuc.msg; then
     remote_connect
  fi
  if grep -q History ./usuc.msg; then
     TAILLINES="100"
     echo "Channel \"$CHANNEL\" History"
     echo "---------------------------------------------"
     echo
     else
     TAILLINES="20"
  fi
  if grep -q DESTROYING ./usuc.msg; then
     echo "Destroying \"$CHANNEL\" channel chat archive"
     echo "---------------------------------------------"
     echo
     rm ./usuc.msg
     break
  fi
  rm ./usuc.msg
done

# Destroy chat file 
# ========================

srm -v ./chat/$CHANNEL.out
srm -v ./chat/$CHANNEL.in
srm -v ./chat/$CHANNEL
echo
if [ -f ./chat/$CHANNEL.out ] || [ -f ./chat/$CHANNEL.in ] || [ -f ./chat/$CHANNEL ]
then
    echo "!!! FAILED to destroy chat archive !!!"
    echo "Destroy \"$CHANNEL\" chat archive manually"
else
    echo "---------------------------------------------"
    echo "CLEAN: \"$CHANNEL\" chat archive destroyed"
    echo
fi

# End
# =========================

exit 1

